"""Docking workflow."""

from typing import Annotated, Self, TypeAlias

from pydantic import AfterValidator, ConfigDict, field_validator, model_validator

from ..base import Base, round_float
from ..pdb import PDB
from ..types import UUID, Vector3D
from .conformer_search import ConformerGenSettingsUnion, ETKDGSettings
from .workflow import MoleculeWorkflow

ProteinUUID: TypeAlias = UUID
CalculationUUID: TypeAlias = UUID


class Score(Base):
    """
    Pose with its score.

    :param pose: conformation of the ligand when docked (calculation UUID)
    :param complex_pdb: the UUID of the protein–ligand complex (protein UUID)
    :param score: score of the pose, in kcal/mol
    :param posebusters_valid: whether or not the ligand pose passes the PoseBusters tests
    :param strain: strain in kcal/mol
    """

    pose: CalculationUUID | None
    complex_pdb: ProteinUUID | None
    score: Annotated[float, AfterValidator(round_float(3))]
    posebusters_valid: bool
    strain: float | None


class DockingSettings(Base):
    """
    Base class for controlling how docked poses are generated.

    :param max_poses: the maximum number of poses generated per input molecule
    """

    max_poses: int = 4


class VinaSettings(DockingSettings):
    """
    Controls how AutoDock Vina is run.

    :param exhaustiveness: how many times Vina attempts to find a pose.
        8 is typical, 32 is considered relatively careful.
    """

    exhaustiveness: int = 8


class DockingWorkflow(MoleculeWorkflow):
    """
    Docking workflow.

    Note that the protein can be supplied either by UUID or raw PDB object.
    We anticipate that the former will dominate deployed usage, but the latter is handy for isolated testing.
    If, for whatever reason, the workflow is initialized with both a `target_uuid` and a `target`, the UUID will be ignored.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (currently unused)

    New:
    :param do_csearch: whether to csearch starting structures
    :param conformer_gen_settings: settings for initial conformer search.
    :param do_optimization: whether to optimize starting structures
    :param optimization_settings: settings for conformer optimization.
    :param do_pose_refinement: whether to optimize non-rotatable bonds in output poses
    :param target: PDB of the protein.
    :param target_uuid: UUID of the protein.
    :param pocket: center (x, y, z) and size (x, y, z) of the pocket

    Results:
    :param conformers: UUIDs of optimized conformers
    :param scores: docked poses sorted by score
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    target: PDB | None = None
    target_uuid: UUID | None = None
    pocket: tuple[Vector3D, Vector3D]

    docking_settings: VinaSettings = VinaSettings()

    do_csearch: bool = True
    conformer_gen_settings: ConformerGenSettingsUnion = ETKDGSettings(mode="reckless")
    do_optimization: bool = True
    do_pose_refinement: bool = True

    conformers: list[CalculationUUID] = []
    scores: list[Score] = []

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        """Return a string representation of the Docking workflow."""
        if self.target is not None:
            desc = self.target.description
            target = desc.code or desc.title
        else:
            target = ""

        ligand = "".join(atom.atomic_symbol for atom in self.initial_molecule.atoms)
        return f"<{type(self).__name__} {target} {ligand}>"

    @model_validator(mode="after")
    def check_protein(self) -> Self:
        """Check if protein is provided."""
        if not self.target and not self.target_uuid:
            raise ValueError("Must provide either target or target_uuid")
        return self

    @field_validator("pocket", mode="after")
    def validate_pocket(cls, pocket: tuple[Vector3D, Vector3D]) -> tuple[Vector3D, Vector3D]:
        _center, size = pocket
        if any(q <= 0 for q in size):
            raise ValueError(f"Pocket size must be positive, got: {size}")
        return pocket



from typing import Annotated, TypeAlias

import numpy as np
import pydantic

from .base import Base
from .types import Matrix3x3, round_matrix3x3

Bool3: TypeAlias = tuple[bool, bool, bool]


class PeriodicCell(Base):
    lattice_vectors: Annotated[Matrix3x3, pydantic.AfterValidator(round_matrix3x3(6))]
    is_periodic: Bool3 = (True, True, True)

    @pydantic.field_validator("lattice_vectors")
    @classmethod
    def check_tensor_3D(cls, v: Matrix3x3) -> Matrix3x3:
        if len(v) != 3 or any(len(row) != 3 for row in v):
            raise ValueError("Cell tensor must be a 3x3 list of floats")

        return v

    @pydantic.field_validator("is_periodic")
    @classmethod
    def check_pbc(cls, v: Bool3) -> Bool3:
        if not any(v):
            raise ValueError("For periodic boundary conditions, at least one dimension must be periodic!")
        return v

    @pydantic.computed_field  # type: ignore[misc, prop-decorator, unused-ignore]
    @property
    def volume(self) -> float:
        return float(np.abs(np.linalg.det(np.array(self.lattice_vectors))))



from pytest import raises

from stjames import Constraint


def test_invalid_constraint_settings() -> None:
    Constraint(constraint_type="bond", atoms=(1, 2))

    with raises(ValueError):
        Constraint(constraint_type="bond", atoms=(1, 2, 3))

    with raises(ValueError):
        Constraint(constraint_type="bond", atoms=(1, 2, 3, 4))

    Constraint(constraint_type="angle", atoms=(1, 2, 3))

    with raises(ValueError):
        Constraint(constraint_type="angle", atoms=(1, 2))

    with raises(ValueError):
        Constraint(constraint_type="angle", atoms=(1, 2, 3, 4))

    Constraint(constraint_type="dihedral", atoms=(1, 2, 3, 4))

    with raises(ValueError):
        Constraint(constraint_type="dihedral", atoms=(1, 2))

    with raises(ValueError):
        Constraint(constraint_type="dihedral", atoms=(1, 2, 3))



"""Nuclear-magnetic-resonance spectroscopy workflow."""

from typing import Annotated

from pydantic import AfterValidator

from ..base import Base, LowercaseStrEnum, round_float
from ..mode import Mode
from ..settings import Settings
from ..solvent import Solvent
from ..types import UUID, round_list
from .conformer_search import ConformerGenSettingsUnion, iMTDSettings
from .multistage_opt import MultiStageOptSettings
from .workflow import MoleculeWorkflow


class NMRMethod(LowercaseStrEnum):
    MAGNETZERO = "magnet-zero"


class NMRPeak(Base):
    """
    Represents a single NMR peak.

    :param nucleus: the atomic number of the nucleus in question
    :param shift: the chemical shift of the peak
    :param atom_indices: the zero-indices of the atoms giving rise to the peak
    """

    nucleus: int
    shift: Annotated[float, AfterValidator(round_float(3))]
    atom_indices: list[int]


class NMRSpectroscopyWorkflow(MoleculeWorkflow):
    """
    Workflow for calculating NMR spectra.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (currently unused)

    New:
    :param nmr_method: how to run the NMR calculations
    :param solvent: the solvent in which to run the calculations
    :param conf_gen_settings : the conformer-search settings. if `None`, no conformer search will be performed
    :param multistage_opt_settings: the optimization settings. if `None`, no optimization will be performed

    Results:
    :param conformers: list of conformer UUIDs
    :param boltzmann_weights: the boltzmann weights for each conformer
    :param per_conformer_chemical_shifts: the per-atom shifts for each conformer
    :param chemical_shifts: the per-atom shifts
    :param symmetry_equivalent_nuclei: 0-indexed atoms which are equivalent to one another
    :param predicted_peaks: the predicted NMR peaks
    """

    nmr_method: NMRMethod = NMRMethod.MAGNETZERO
    solvent: Solvent = Solvent.CHLOROFORM

    conf_gen_settings: ConformerGenSettingsUnion | None = iMTDSettings(mode="careful")
    multistage_opt_settings: MultiStageOptSettings | None = MultiStageOptSettings(
        mode=Mode.MANUAL,
        optimization_settings=[Settings(method="aimnet2_wb97md3", tasks=["optimize"])],
    )

    conformers: list[UUID] = []
    boltzmann_weights: Annotated[list[float], AfterValidator(round_list(3))] = []
    per_conformer_chemical_shifts: list[Annotated[list[float | None], AfterValidator(round_list(3))]] = []
    chemical_shifts: Annotated[list[float | None], AfterValidator(round_list(3))] = []
    symmetry_equivalent_nuclei: list[list[int]] = []

    predicted_peaks: dict[int, list[NMRPeak]] = {}



from pytest import mark

from stjames.pdb import (
    PDB,
    PDBDescription,
    PDBExperiment,
    PDBModel,
    fetch_pdb,
    fetch_pdb_from_mmcif,
    pdb_from_mmcif_filestring,
    pdb_from_pdb_filestring,
    pdb_object_to_pdb_filestring,
)


def test_1ema() -> None:
    """Green fluorescent protein."""
    fetch_pdb("1EMA")


def test_8qvy() -> None:
    """
    Human GABAA receptor
    Not availble as a .pdb
    """
    fetch_pdb_from_mmcif("8VQY")


def test_read_pdb_filestring() -> None:
    """Rest reading of a pdb string."""
    with open("tests/data/1ema.pdb") as f:
        data = f.read()
    pdb = pdb_from_pdb_filestring(data)
    assert pdb.description.code == "1EMA"

    json = pdb.model_dump()
    PDB.model_validate(json)


def test_read_pdb_filestring_mome() -> None:
    """Rest reading of a pdb string."""
    with open("tests/data/cluster_1.pdb") as f:
        data = f.read()
    pdb = pdb_from_pdb_filestring(data)

    json = pdb.model_dump()
    print(json)
    PDB.model_validate(json)


def test_read_mmcif_filestring() -> None:
    """Rest reading of a mmcif string."""
    with open("tests/data/1ema.cif") as f:
        data = f.read()
    pdb = pdb_from_mmcif_filestring(data)
    assert pdb.description.code == "1EMA"

    json = pdb.model_dump()
    PDB.model_validate(json)


# fmt: off
@mark.regression
@mark.parametrize(
    "code",
    [
        # Codes from molecules of the month August 2024–January 2025
        "7S6B", "8UCS", "2ZVY", "1F4V", "6YKM", "6E10", "6E11", "3VCM", "2X0B",
        "6OS0", "1N9U", "1O86", "2V0Z", "6KI1", "6KI2", "7YYO", "6TJV", "7EGL",
        "7CYF", "7EGK", "7ZCG", "3FRT", "6AP1",
        # Codes created by o1
        "1CRN", "1MBN", "4HHB", "1HHO", "1BNA", "1CAG", "2JHO", "1EVV", "3ZOJ",
        "4AGG", "2Y69", "6R1V", "6ND2", "7NZ6", "1S72", "3G5U", "7DFT", "6AI0",
        "6NG2", "1A0I", "1B7C", "1C8R", "1D4T", "1E7O", "1F9J", "1G5K", "1H8L",
        "1I2M", "1J3N", "1K4P", "1M7R", "1N8S", "1O9T", "1P0U", "1Q1V", "1R2W",
        "1S3X", "1T4Y", "1U5Z", "1V6A", "1W7B", "1X8C", "1Y9D", "1Z0E", "2A1F",
        "2B2G", "2C3H", "2D4I", "2E5J", "2F6K", "2G7L", "2H8M", "2I9N", "2J0O",
        "2K1P", "2L2Q", "2N4S", "2O5T", "2P6U", "2Q7V", "2R8W", "2V2A", "2W3B",
        "2X4C", "2Y5D", "2Z6E", "3A7F", "3B8G", "3C9H", "3D0I", "3E1J", "3F2K",
        "3G3L", "3H4M", "3I5N", "3J6O", "3K7P", "3L8Q", "3N0S", "3O1T", "3P2U",
        "3Q3V", "3S5X", "3T6Y", "3U7Z", "3V8A", "3W9B", "3X0C", "4A3F", "4B4G",
        "4C5H", "4D6I", "4E7J", "4F8K", "4G9L", "4H0M", "4I1N", "4J2O", "4K3P",
        "4L4Q", "4M5R", "4N6S", "4O7T", "4P8U", "4Q9V", "4R0W", "4S1X",
    ]
)  # fmt: on
def test_pdb(code: str) -> None:
    pdb = fetch_pdb(code)

    json = pdb.model_dump()
    PDB.model_validate(json)

def test_from_pdb_to_pdb_2qto() -> None:
    with open("tests/data/2qto.pdb") as f:
        data = f.read()
    pdb = pdb_from_pdb_filestring(data)
    filestring = pdb_object_to_pdb_filestring(pdb, header=True, source=True, keyword=True, crystallography=True)
    pdb2 = pdb_from_pdb_filestring(filestring)

    assert pdb.description == pdb2.description
    assert pdb.experiment == pdb2.experiment
    assert pdb.models == pdb2.models

def test_from_pdb_to_pdb_1ema() -> None:
    with open("tests/data/1ema.pdb") as f:
        data = f.read()
    pdb = pdb_from_pdb_filestring(data)
    filestring = pdb_object_to_pdb_filestring(pdb, header=True, source=True, keyword=True, crystallography=True)
    pdb2 = pdb_from_pdb_filestring(filestring)

    assert pdb.description == pdb2.description
    assert pdb.models == pdb2.models

def test_from_pdb_to_pdb_2hu4() -> None:
    with open("tests/data/2HU4.pdb") as f:
        data = f.read()
    pdb = pdb_from_pdb_filestring(data)
    filestring = pdb_object_to_pdb_filestring(pdb, header=True, source=True, keyword=True, crystallography=True)

    pdb2 = pdb_from_pdb_filestring(filestring)

    assert pdb.description == pdb2.description
    # not true but doesn't matter
    print(pdb.geometry == pdb2.geometry)
    assert pdb.models == pdb2.models

def mmcif_author_format_to_pdb_format(authors: list[str]) -> list[str]:
    return [f"{last.upper()}{first.upper()}" for first, last in
            (author.split(", ") for author in authors)]

def compare_descriptions_mmcif_and_pdb(mmcif_description: PDBDescription, pdb_description: PDBDescription) -> bool:
    return (
        mmcif_description.code == pdb_description.code and
        mmcif_description.title == pdb_description.title and
        mmcif_author_format_to_pdb_format(mmcif_description.authors) == (pdb_description.authors) and
        mmcif_description.classification == pdb_description.classification and
        mmcif_description.deposition_date == pdb_description.deposition_date and
        sorted(mmcif_description.keywords) == sorted(pdb_description.keywords)
    )

def compare_experiments_mmcif_and_pdb(mmcif_experiment: PDBExperiment, pdb_experiment: PDBExperiment) -> bool:
    return (
        mmcif_experiment.expression_system.upper() == pdb_experiment.expression_system and  # type: ignore [union-attr]
        mmcif_experiment.missing_residues == pdb_experiment.missing_residues and
        mmcif_experiment.source_organism.upper() == pdb_experiment.source_organism and  # type: ignore [union-attr]
        mmcif_experiment.technique == pdb_experiment.technique
    )

def compare_models_mmcif_and_pdb(mmcif_models: list[PDBModel], pdb_models: list[PDBModel]) -> bool:
    for i in range(len(mmcif_models)):
        assert mmcif_models[i].polymer == pdb_models[i].polymer
        assert mmcif_models[i].non_polymer == pdb_models[i].non_polymer
        assert mmcif_models[i].branched == pdb_models[i].branched
    return True

def test_from_mmcif_to_pdb_1ema() -> None:
    with open("tests/data/1ema.cif") as f:
        mmcif_data = f.read()

    with open("tests/data/1ema.pdb") as f:
        pdb_data = f.read()
    mmcif_1ema = pdb_from_mmcif_filestring(mmcif_data)
    pdb_1ema = pdb_from_pdb_filestring(pdb_data)

    assert compare_descriptions_mmcif_and_pdb(mmcif_1ema.description, pdb_1ema.description)
    assert compare_experiments_mmcif_and_pdb(mmcif_1ema.experiment, pdb_1ema.experiment)
    assert mmcif_1ema.quality == pdb_1ema.quality
    assert compare_models_mmcif_and_pdb(mmcif_1ema.models, pdb_1ema.models)



"""Fukui index workflow."""

from typing import Annotated, Self

from pydantic import AfterValidator, model_validator

from stjames.method import Method

from ..base import round_optional_float
from ..engine import Engine
from ..settings import Settings
from ..types import UUID, FloatPerAtom, round_optional_float_per_atom
from .workflow import MoleculeWorkflow


class FukuiIndexWorkflow(MoleculeWorkflow):
    """
    Workflow for calculating Fukui indices.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (currently unused)

    Settings:
    :param opt_settings: if given, the settings for optimization. if none, no optimization will be conducted.
    :param opt_engine: the engine for optimization [uses opt_settings.method.default_engine if not set]
    :param fukui_settings: the settings for Fukui index calculations.
    :param fukui_engine: the engine for Fukui index calculations [uses fukui_settings.method.default_engine if not set]

    Results:
    :param optimization: UUID of optimization
    :param global_electrophilicity_index: global electrophilicity index
    :param fukui_positive: Fukui index for positive charges
    :param fukui_negative: Fukui index for negative charges
    :param fukui_zero: Fukui index for zero charges
    """

    opt_settings: Settings | None = None
    opt_engine: Engine | None = None

    fukui_settings: Settings = Settings(method=Method.GFN1_XTB)
    fukui_engine: Engine = None  # type: ignore [assignment]

    optimization: UUID | None = None

    global_electrophilicity_index: Annotated[float | None, AfterValidator(round_optional_float(6))] = None
    fukui_positive: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    fukui_negative: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    fukui_zero: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None

    @model_validator(mode="after")
    def set_engines(self) -> Self:
        """Set the engines for optimization and Fukui index calculations."""
        if self.opt_settings is not None and self.opt_engine is None:
            self.opt_engine = self.opt_settings.method.default_engine()
        self.fukui_engine = self.fukui_engine or self.fukui_settings.method.default_engine()

        return self



import re
from pathlib import Path
from typing import Annotated, Any, Iterable, Optional, Self, Sequence, TypeAlias, TypedDict, TypeVar

import numpy as np
import pydantic
from pydantic import AfterValidator, NonNegativeInt, PositiveInt, ValidationError
from rdkit import Chem
from rdkit.Chem import AllChem

from .atom import Atom
from .base import Base, round_float, round_optional_float
from .data import SYMBOL_ELEMENT
from .periodic_cell import PeriodicCell
from .types import (
    FloatPerAtom,
    Matrix3x3,
    Vector3D,
    Vector3DPerAtom,
    round_optional_float_per_atom,
    round_optional_matrix3x3,
    round_optional_vector3d,
    round_optional_vector3d_per_atom,
    round_vector3d_per_atom,
)

RdkitMol: TypeAlias = Chem.rdchem.Mol | Chem.rdchem.RWMol


class MoleculeReadError(RuntimeError):
    pass


class VibrationalMode(Base):
    frequency: Annotated[float, AfterValidator(round_float(3))]  # in cm-1
    reduced_mass: Annotated[float, AfterValidator(round_float(3))]  # amu
    force_constant: Annotated[float, AfterValidator(round_float(3))]  # mDyne/Å
    displacements: Annotated[Vector3DPerAtom, AfterValidator(round_vector3d_per_atom(6))]  # Å
    ir_intensity: Annotated[Optional[float], AfterValidator(round_optional_float(3))] = None  # km/mol


class Molecule(Base):
    charge: int
    multiplicity: PositiveInt
    atoms: list[Atom]

    # for periodic boundary conditions
    cell: Optional[PeriodicCell] = None

    energy: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None  # in Hartree
    scf_iterations: Optional[NonNegativeInt] = None
    scf_completed: Optional[bool] = None
    elapsed: Annotated[Optional[float], AfterValidator(round_optional_float(3))] = None  # in seconds

    homo_lumo_gap: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None  # in eV

    gradient: Annotated[Optional[Vector3DPerAtom], AfterValidator(round_optional_vector3d_per_atom(6))] = None  # Hartree/Å
    stress: Annotated[Optional[Matrix3x3], AfterValidator(round_optional_matrix3x3(6))] = None  # Hartree/Å

    velocities: Annotated[Optional[Vector3DPerAtom], AfterValidator(round_optional_vector3d_per_atom(6))] = None  # Å/fs

    mulliken_charges: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    mulliken_spin_densities: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    dipole: Annotated[Optional[Vector3D], AfterValidator(round_optional_vector3d(6))] = None  # in Debye

    vibrational_modes: Optional[list[VibrationalMode]] = None

    zero_point_energy: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None
    thermal_energy_corr: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None
    thermal_enthalpy_corr: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None
    thermal_free_energy_corr: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None

    smiles: Optional[str] = None
    calculation_index: int | None = None

    def __len__(self) -> int:
        return len(self.atoms)

    def distance(self, i: PositiveInt, j: PositiveInt) -> float:
        r"""
        Calculate the distance between atoms.

        >>> mol = Molecule.from_xyz("H 0 1 0\nH 0 0 1")
        >>> mol.distance(1, 2)
        1.4142135623730951
        """
        return sum((q2 - q1) ** 2 for q1, q2 in zip(self.atoms[i - 1].position, self.atoms[j - 1].position)) ** 0.5  # type: ignore [no-any-return,unused-ignore]

    def angle(self, i: PositiveInt, j: PositiveInt, k: PositiveInt, degrees: bool = True) -> float:
        r"""
        Calculate the angle between three atoms.

        >>> Molecule.from_xyz("H 0 0 0\nO 0 0 1\nH 0 1 1").angle(1, 2, 3)
        90.0
        """

        return angle(self.coordinates[i - 1], self.coordinates[j - 1], self.coordinates[k - 1], degrees=degrees)

    def dihedral(self, i: int, j: int, k: int, l: int, degrees: bool = True, positive_domain: bool = True) -> float:
        r"""
        Calculate the dihedral angle between four atoms.

        >>> Molecule.from_xyz("H 0 0 0\nO 0 0 1\nO 0 1 1\nH 1 1 1").dihedral(1, 2, 3, 4)
        270.0
        """
        return dihedral(
            self.coordinates[i - 1],
            self.coordinates[j - 1],
            self.coordinates[k - 1],
            self.coordinates[l - 1],
            degrees=degrees,
            positive_domain=positive_domain,
        )

    @property
    def coordinates(self) -> Vector3DPerAtom:
        return [a.position for a in self.atoms]

    def translated(self, vector: Vector3D) -> Self:
        r"""
        Translate the molecule by a vector.

        >>> mol = Molecule.from_xyz("H 0 0 0\nH 0 0 1")
        >>> print(mol.translated((1, 0, 0)).to_xyz())
        2
        <BLANKLINE>
        H     1.0000000000    0.0000000000    0.0000000000
        H     1.0000000000    0.0000000000    1.0000000000
        """

        def translated(position: Vector3D) -> Vector3D:
            return tuple(q + v for q, v in zip(position, vector, strict=True))  # type: ignore [return-value]

        atoms = [atom.copy(update={"position": translated(atom.position)}) for atom in self.atoms]

        return self.copy(update={"atoms": atoms})

    @property
    def atomic_numbers(self) -> list[NonNegativeInt]:
        return [a.atomic_number for a in self.atoms]

    @property
    def sum_energy_zpe(self) -> Optional[float]:
        if (self.energy is None) or (self.zero_point_energy is None):
            return None
        return self.energy + self.zero_point_energy

    @property
    def sum_energy_thermal_corr(self) -> Optional[float]:
        if (self.energy is None) or (self.thermal_energy_corr is None):
            return None
        return self.energy + self.thermal_energy_corr

    @property
    def sum_energy_enthalpy(self) -> Optional[float]:
        if (self.energy is None) or (self.thermal_enthalpy_corr is None):
            return None
        return self.energy + self.thermal_enthalpy_corr

    @property
    def enthalpy(self) -> Optional[float]:
        return self.sum_energy_enthalpy

    @property
    def sum_energy_free_energy(self) -> Optional[float]:
        if (self.energy is None) or (self.thermal_free_energy_corr is None):
            return None
        return self.energy + self.thermal_free_energy_corr

    @property
    def gibbs_free_energy(self) -> Optional[float]:
        return self.sum_energy_free_energy

    @pydantic.model_validator(mode="after")
    def check_electron_sanity(self) -> Self:
        num_electrons = sum(self.atomic_numbers) - self.charge
        num_unpaired_electrons = self.multiplicity - 1
        if (num_electrons - num_unpaired_electrons) % 2 != 0:
            raise ValueError(
                f"The combination of {num_electrons} electrons, charge {self.charge}, and multiplicity {self.multiplicity} is impossible. "
                "Double-check the charge and multiplicity values given and verify that they are correct."
            )

        return self

    @classmethod
    def from_file(cls: type[Self], filename: Path | str, format: str | None = None, charge: int = 0, multiplicity: PositiveInt = 1) -> Self:
        r"""
        Read a molecule from a file.

        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile("w+", suffix=".xyz") as f:
        ...    _ = f.write("2\nComment\nH 0 0 0\nF 0 0 1")
        ...    _ = f.seek(0)
        ...    mol = Molecule.from_file(f.name)
        >>> print(mol.to_xyz())
        2
        <BLANKLINE>
        H     0.0000000000    0.0000000000    0.0000000000
        F     0.0000000000    0.0000000000    1.0000000000
        """
        filename = Path(filename)
        if not format:
            format = filename.suffix[1:]

        with open(filename) as f:
            match format:
                case "xyz":
                    return cls.from_xyz_lines(f.readlines(), charge=charge, multiplicity=multiplicity)
                case "extxyz":
                    return cls.from_extxyz_lines(f.readlines(), charge=charge, multiplicity=multiplicity)
                case _:
                    raise ValueError(f"Unsupported {format=}")

    @classmethod
    def from_xyz(cls: type[Self], xyz: str, charge: int = 0, multiplicity: PositiveInt = 1) -> Self:
        r"""
        Generate a Molecule from an XYZ string.

        Note: only supports single molecule inputs.

        >>> len(Molecule.from_xyz("2\nComment\nH 0 0 0\nH 0 0 1"))
        2
        """
        return cls.from_xyz_lines(xyz.strip().splitlines(), charge=charge, multiplicity=multiplicity)

    @classmethod
    def from_xyz_lines(cls: type[Self], lines: Iterable[str], charge: int = 0, multiplicity: PositiveInt = 1) -> Self:
        lines = list(lines)
        if len(lines[0].split()) == 1:
            natoms = lines[0].strip()
            if not natoms.isdigit() or (int(lines[0]) != len(lines) - 2):
                raise MoleculeReadError(f"First line of XYZ file should be the number of atoms, got: {lines[0]} != {len(lines) - 2}")
            lines = lines[2:]

        try:
            return cls(atoms=[Atom.from_xyz(line) for line in lines], charge=charge, multiplicity=multiplicity)
        except (ValueError, ValidationError) as e:
            raise MoleculeReadError("Error reading molecule from xyz") from e

    def to_xyz(self, comment: str = "", out_file: Path | str | None = None) -> str:
        r"""
        Generate an XYZ string.

        >>> mol = Molecule.from_xyz("2\nComment\nH 0 1 2\nF 1 2 3")
        >>> print(mol.to_xyz(comment="HF"))
        2
        HF
        H     0.0000000000    1.0000000000    2.0000000000
        F     1.0000000000    2.0000000000    3.0000000000
        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as directory:
        ...     file = Path(directory) / "mol.xyz"
        ...     out = mol.to_xyz(comment="HF", out_file=file)
        ...     with file.open() as f:
        ...         Molecule.from_xyz(f.read()).to_xyz("HF") == out
        True
        """
        geom = "\n".join(map(str, self.atoms))
        out = f"{len(self)}\n{comment}\n{geom}"

        if out_file:
            with Path(out_file).open("w") as f:
                f.write(out)

        return out

    @classmethod
    def from_extxyz(cls: type[Self], extxyz: str, charge: int = 0, multiplicity: PositiveInt = 1) -> Self:
        r"""
        Generate a Molecule from a EXTXYZ string. Currently only supporting Lattice and Properties fields.

        >>> Molecule.from_extxyz('''
        ... 2
        ... Lattice="6.0 0.0 0.0 6.0 0.0 0.0 6.0 0.0 0.0"Properties=species:S:1:pos:R:3
        ... H 0 0 0
        ... H 0 0 1
        ... ''').cell.lattice_vectors
        ((6.0, 0.0, 0.0), (6.0, 0.0, 0.0), (6.0, 0.0, 0.0))
        """

        return cls.from_extxyz_lines(extxyz.strip().splitlines(), charge=charge, multiplicity=multiplicity)

    @classmethod
    def from_extxyz_lines(
        cls: type[Self],
        lines: Iterable[str],
        charge: int | None = None,
        multiplicity: PositiveInt | None = None,
        cell: PeriodicCell | None = None,
    ) -> Self:
        """
        Parses an EXTXYZ file, extracting atom positions, forces (if present), and metadata.

        Supports:
        - Lattice vectors (cell information)
        - Properties field (species, positions, forces, etc.)
        - Other metadata like charge, multiplicity, energy, etc.

        :param lines: Iterable of lines from an EXTXYZ file
        :param charge: total charge of the molecule (default: 0 if not found)
        :param multiplicity: spin multiplicity of the molecule (default: 1 if not found)
        :param cell: PeriodicCell containing lattice vectors
        :return: Molecule
        :raises MoleculeReadError: if the file is not in the correct format
        """
        if not isinstance(lines, Sequence):
            lines = list(lines)

        # Ensure first line contains number of atoms
        if len(lines[0].split()) == 1:
            natoms = lines[0].strip()
            if not natoms.isdigit() or (int(natoms) != len(lines) - 2):
                raise MoleculeReadError(f"First line should be number of atoms, got: {lines[0]} != {len(lines) - 2}")
            data_line, *lines = lines[1:]
        else:
            raise MoleculeReadError(f"First line should be an integer denoting atom count. Got {lines[0].split()}")

        metadata = parse_extxyz_comment_line(data_line)

        T = TypeVar("T")

        def metadata_optional_get(key: str, value: T | None, default: T) -> T:
            """Set key to default if not found in metadata"""
            if value is None:
                return metadata.get(key, default)  # type: ignore [return-value]

            return value

        charge = metadata_optional_get("total_charge", charge, 0)
        multiplicity = metadata_optional_get("multiplicity", multiplicity, 1)
        cell = cell or metadata.get("cell")
        energy = metadata.get("energy", None)

        force_idx = None
        if properties := metadata.get("properties", "").split(":"):
            if properties[0].lower() != "species":
                raise MoleculeReadError(f"Invalid or missing 'Properties' field in EXTXYZ, got: {properties}")

            # Identify column indices for position and force data
            pos_idx = None
            current_idx = 0  # Start after 'species:S'

            while current_idx < len(properties):
                if properties[current_idx].lower() == "pos" and properties[current_idx + 1].lower() == "r" and properties[current_idx + 2] == "3":
                    pos_idx = current_idx
                elif properties[current_idx].lower() == "forces" and properties[current_idx + 1].lower() == "r" and properties[current_idx + 2] == "3":
                    force_idx = current_idx
                current_idx += 3

            if pos_idx is None:
                raise MoleculeReadError("No position data ('pos:R:3') found in Properties field.")

        def parse_line_atoms(line: str) -> Atom:
            symbol, sx, sy, sz, *_ = line.split()
            atomic_number = SYMBOL_ELEMENT[symbol.title()]
            x, y, z = map(float, (sx, sy, sz))

            return Atom(atomic_number=atomic_number, position=(x, y, z))

        def parse_line_with_grad(line: str) -> tuple[Atom, Vector3D]:
            symbol, sx, sy, sz, sgx, sgy, sgz, *_ = line.split()
            atomic_number = SYMBOL_ELEMENT[symbol.title()]
            x, y, z = map(float, (sx, sy, sz))
            gx, gy, gz = map(float, (sgx, sgy, sgz))

            return (
                Atom(atomic_number=atomic_number, position=(x, y, z)),
                (-gx, -gy, -gz),
            )

        atoms: list[Atom]
        gradients: list[Vector3D] | None
        if force_idx is not None:
            atoms, gradients = zip(*map(parse_line_with_grad, lines), strict=True)  # type: ignore [assignment]
        else:
            atoms = [parse_line_atoms(line) for line in lines]
            gradients = None

        return cls(atoms=atoms, cell=cell, charge=charge, multiplicity=multiplicity, energy=energy, gradient=gradients)

    @classmethod
    def from_rdkit(cls: type[Self], rdkm: RdkitMol, cid: int = 0, multiplicity: int = 1) -> Self:
        if len(rdkm.GetConformers()) == 0:
            rdkm = _embed_rdkit_mol(rdkm)

        atomic_numbers = [atom.GetAtomicNum() for atom in rdkm.GetAtoms()]  # type: ignore [no-untyped-call, unused-ignore]
        atoms = [
            Atom(atomic_number=atom, position=xyz)  # keep open
            for atom, xyz in zip(atomic_numbers, rdkm.GetConformers()[cid].GetPositions(), strict=True)
        ]

        charge = Chem.GetFormalCharge(rdkm)

        return cls(atoms=atoms, charge=charge, multiplicity=multiplicity)

    @classmethod
    def from_smiles(cls: type[Self], smiles: str) -> Self:
        rdkm = Chem.MolFromSmiles(smiles)
        assert rdkm is not None
        return cls.from_rdkit(rdkm)


def _embed_rdkit_mol(rdkm: RdkitMol) -> RdkitMol:
    try:
        AllChem.SanitizeMol(rdkm)  # type: ignore [attr-defined, unused-ignore]
    except Exception as e:
        raise ValueError("Molecule could not be generated -- invalid chemistry!\n") from e

    rdkm = AllChem.AddHs(rdkm)  # type: ignore [attr-defined, unused-ignore]
    try:
        status1 = AllChem.EmbedMolecule(rdkm, maxAttempts=200)  # type: ignore [attr-defined, unused-ignore]
        assert status1 >= 0
    except Exception as e:
        status1 = AllChem.EmbedMolecule(rdkm, maxAttempts=200, useRandomCoords=True)  # type: ignore [attr-defined, unused-ignore]
        if status1 < 0:
            raise ValueError(f"Cannot embed molecule! Error: {e}")

    AllChem.MMFFOptimizeMolecule(rdkm, maxIters=200)  # type: ignore [attr-defined, call-arg, unused-ignore]

    return rdkm


class EXTXYZMetadata(TypedDict, total=False):
    properties: Any
    total_charge: int
    multiplicity: int
    energy: float
    cell: PeriodicCell


def parse_extxyz_comment_line(line: str) -> EXTXYZMetadata:
    """
    Parse the comment line of an EXTXYZ file, extracting lattice, properties, and metadata.

    Supports:
    - Lattice vectors (cell information)
    - Properties field (species, positions, forces, etc.)
    - Other metadata fields like charge, multiplicity, energy, etc.

    :param line: comment line from an EXTXYZ file
    :return: parsed properties

    >>> parse_extxyz_comment_line('Lattice="6.0 0.0 0.0 6.0 0.0 0.0 6.0 0.0 0.0"Properties=species:S:1:pos:R:3')
    {'cell': PeriodicCell(lattice_vectors=((6.0, 0.0, 0.0), (6.0, 0.0, 0.0), (6.0, 0.0, 0.0)), is_periodic=(True, True, True), volume=0.0), 'properties': 'species:S:1:pos:R:3'}
    """  # noqa: E501

    # Regular expression to match key="value", key='value', or key=value
    pattern = r"(\S+?=(?:\".*?\"|\'.*?\'|\S+))"
    pairs = re.findall(pattern, line)

    prop_dict: EXTXYZMetadata = {}
    for pair in pairs:
        key, value = pair.split("=", 1)
        key = key.lower().strip()
        value = value.strip("'\"")

        if key == "lattice":
            lattice_values = value.split()
            if len(lattice_values) != 9:
                raise MoleculeReadError(f"Lattice should have 9 entries, got {len(lattice_values)}")

            try:
                cell = tuple(tuple(map(float, lattice_values[i : i + 3])) for i in range(0, 9, 3))
            except ValueError:
                raise MoleculeReadError(f"Lattice should be floats, got {lattice_values}")

            prop_dict["cell"] = PeriodicCell(lattice_vectors=cell)

        elif key == "properties":
            prop_dict["properties"] = value

        elif key == "total_charge":
            prop_dict["total_charge"] = int(value)
        elif key == "multiplicity":
            prop_dict["multiplicity"] = int(value)
        elif key == "energy":
            prop_dict["energy"] = float(value)
        else:
            prop_dict[key] = value  # type: ignore [literal-required]

    return prop_dict


def angle(p0: Sequence[float], p1: Sequence[float], p2: Sequence[float], degrees: bool = True) -> float:
    """
    Angle between three points.

    :param i, j, k: positions of points
    :param degrees: whether to return in degrees
    :return: angle in radians or degrees
    """
    a0, a1, a2 = map(np.asarray, (p0, p1, p2))
    u = a1 - a0
    v = a1 - a2

    nu = np.linalg.norm(u)
    nv = np.linalg.norm(v)
    cos_theta = np.dot(u, v) / (nu * nv)
    cos_theta = np.clip(cos_theta, -1.0, 1.0)

    ang = np.arccos(cos_theta)
    if degrees:
        return float(np.degrees(ang))
    return float(ang)


def dihedral(p0: Sequence[float], p1: Sequence[float], p2: Sequence[float], p3: Sequence[float], degrees: bool = True, positive_domain: bool = True) -> float:
    """
    Dihedral angle between four points.

    :param p0, p1, p2, p3: points
    :param degrees: whether to return in degrees
    :param positive_domain: (0, 360] if True else (-180, 180]
    :return: angle in degrees or radians (or nan if collinearities detected)

    >>> a = [0, 0, 0]
    >>> b = [0, 0, 1]
    >>> c = [0, 1, 1]
    >>> d1 = [0, 1, 2]
    >>> d2 = [0.5, 1, 1.5]
    >>> dihedral(a, b, c, d1)
    180.0
    >>> dihedral(a, b, c, d2, positive_domain=False)
    -135.0
    """
    a0, a1, a2, a3 = map(np.asarray, (p0, p1, p2, p3))
    b0 = a1 - a0
    b1 = a2 - a1
    b2 = a3 - a2

    b1 = b1 / np.linalg.norm(b1)

    v = b1 * np.dot(b0, b1) - b0
    w = b2 - b1 * np.dot(b2, b1)

    x = np.dot(v, w)
    y = np.dot(np.cross(b1, v), w)
    ang = np.arctan2(y, x)

    if positive_domain and ang < 0:
        ang += 2 * np.pi

    if degrees:
        return float(np.degrees(ang))
    return float(ang)



# ruff: noqa: F405

from typing import Literal

from .admet import *
from .basic_calculation import *
from .bde import *
from .conformer import *
from .conformer_search import *
from .descriptors import *
from .docking import *
from .double_ended_ts_search import DoubleEndedTSSearchWorkflow
from .electronic_properties import *
from .fukui import *
from .hydrogen_bond_basicity import *
from .ion_mobility import *
from .irc import *
from .macropka import *
from .molecular_dynamics import *
from .multistage_opt import *
from .nmr import *
from .pka import *
from .pose_analysis_md import *
from .protein_cofolding import *
from .redox_potential import *
from .scan import *
from .solubility import *
from .spin_states import *
from .strain import *
from .tautomer import *
from .workflow import *

WORKFLOW_NAME = Literal[
    "admet",
    "basic_calculation",
    "bde",
    "conformers",
    "conformer_search",
    "descriptors",
    "docking",
    "double_ended_ts_search",
    "electronic_properties",
    "fukui",
    "hydrogen_bond_basicity",
    "ion_mobility",
    "irc",
    "macropka",
    "molecular_dynamics",
    "multistage_opt",
    "nmr",
    "pka",
    "pose_analysis_md",
    "protein_cofolding",
    "redox_potential",
    "scan",
    "solubility",
    "spin_states",
    "strain",
    "tautomers",
]

WORKFLOW_MAPPING: dict[WORKFLOW_NAME, Workflow] = {
    "admet": ADMETWorkflow,  # type: ignore [dict-item]
    "basic_calculation": BasicCalculationWorkflow,  # type: ignore [dict-item]
    "bde": BDEWorkflow,  # type: ignore [dict-item]
    "conformers": ConformerWorkflow,  # type: ignore [dict-item]
    "conformer_search": ConformerSearchWorkflow,  # type: ignore [dict-item]
    "descriptors": DescriptorsWorkflow,  # type: ignore [dict-item]
    "docking": DockingWorkflow,  # type: ignore [dict-item]
    "double_ended_ts_search": DoubleEndedTSSearchWorkflow,  # type: ignore [dict-item]
    "electronic_properties": ElectronicPropertiesWorkflow,  # type: ignore [dict-item]
    "fukui": FukuiIndexWorkflow,  # type: ignore [dict-item]
    "hydrogen_bond_basicity": HydrogenBondBasicityWorkflow,  # type: ignore [dict-item]
    "ion_mobility": IonMobilityWorkflow,  # type: ignore [dict-item]
    "irc": IRCWorkflow,  # type: ignore [dict-item]
    "macropka": MacropKaWorkflow,  # type: ignore [dict-item]
    "molecular_dynamics": MolecularDynamicsWorkflow,  # type: ignore [dict-item]
    "multistage_opt": MultiStageOptWorkflow,  # type: ignore [dict-item]
    "nmr": NMRSpectroscopyWorkflow,  # type: ignore [dict-item]
    "pka": pKaWorkflow,  # type: ignore [dict-item]
    "pose_analysis_md": PoseAnalysisMolecularDynamicsWorkflow,  # type: ignore [dict-item]
    "protein_cofolding": ProteinCofoldingWorkflow,  # type: ignore [dict-item]
    "redox_potential": RedoxPotentialWorkflow,  # type: ignore [dict-item]
    "scan": ScanWorkflow,  # type: ignore [dict-item]
    "solubility": SolubilityWorkflow,  # type: ignore [dict-item]
    "spin_states": SpinStatesWorkflow,  # type: ignore [dict-item]
    "strain": StrainWorkflow,  # type: ignore [dict-item]
    "tautomers": TautomerWorkflow,  # type: ignore [dict-item]
}



from typing import Annotated

from pydantic import AfterValidator

from ..base import round_float, round_optional_float
from ..mode import Mode
from ..settings import Settings
from ..types import UUID
from .conformer_search import ConformerGenSettingsUnion, ETKDGSettings
from .multistage_opt import MultiStageOptSettings
from .workflow import MoleculeWorkflow


class StrainWorkflow(MoleculeWorkflow):
    """
    Workflow for calculating the strain of a given molecular geometry.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (currently unused)

    New:
    :param conf_gen_settings : the conformer-search settings.
    :param multistage_opt_settings: the optimization settings.
    :param harmonic_constraint_spring_constant: the spring constant for the constraints in kcal/mol/Å
    :param constrain_hydrogens: whether or not to constrain hydrogens

    Results:
    :param conformers: list of conformer UUIDs
    :param constrained_optimization: the UUID of the optimized strained structure
    :param strain: the actual strain in kcal/mol
    """

    conf_gen_settings: ConformerGenSettingsUnion = ETKDGSettings(mode="rapid")
    multistage_opt_settings: MultiStageOptSettings = MultiStageOptSettings(
        mode=Mode.MANUAL,
        optimization_settings=[Settings(method="aimnet2_wb97md3", tasks=["optimize"])],
        singlepoint_settings=Settings(method="aimnet2_wb97md3", tasks=["energy"], solvent_settings={"solvent": "water", "model": "cpcmx"}),
    )

    harmonic_constraint_spring_constant: Annotated[float, AfterValidator(round_float(3))] = 5.0
    constrain_hydrogens: bool = False

    constrained_optimization: UUID | None = None
    conformers: list[UUID | None] = []
    strain: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
